from collections import UserString
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager
from functools import partial
from typing import Any

import nats
from nats.aio.client import Client as NATS
from nats.js import JetStreamContext

from tfy_assistant_framework._http_client import (
    REQUEST_TIMEOUT,
    async_http_client,
    log_and_raise_for_status,
)
from tfy_assistant_framework._logger import logger
from tfy_assistant_framework.settings import settings

_PING_INTERVAL: int = 6
_MAX_OUTSTANDING_PINGS: int = 5
_MAX_RECONNECT_ATTEMPTS: int = -1
_CONNECTION_TIMEOUT: int = 10

JS_PUBLISH_TIMEOUT: int = 10
JS_TIMEOUT: int = 10


async def _get_nats_credentials() -> str:
    """
    Fetch the NATS user credentials from the TrueFoundry API.
    Returns:
        str: The NATS user credentials.
    """
    if any(
        [
            settings.tfy_base_url is None,
            settings.tfy_api_key is None,
            settings.tfy_assumed_user is None,
        ]
    ):
        raise ValueError("TFY credentials are not set, cannot fetch NATS credentials")
    url = f"{settings.tfy_base_url}/v1/x/llm-agent/nats-creds"
    headers = {
        "Authorization": f"Bearer {settings.tfy_api_key}",
    }
    response = await async_http_client.get(
        url, headers=headers, timeout=REQUEST_TIMEOUT
    )
    log_and_raise_for_status(response)
    data = response.json()
    creds = f"""
-----BEGIN NATS USER JWT-----
{data["jwt"]}
------END NATS USER JWT------
************************* IMPORTANT *************************
Private NKEYs are sensitive and should be treated as secrets.
-----BEGIN USER NKEY SEED-----
{data["seed"]}
------END USER NKEY SEED------
*************************************************************
"""
    return creds


async def _log_nats_connection_event(event: str) -> None:
    logger.warning("NATS connection event: %s", event)


async def _log_nats_error(ex: Exception) -> None:
    logger.error("NATS error", exc_info=ex)


# Define callback functions at module level
_DEFAULT_RECONNECTED_CB = partial(_log_nats_connection_event, event="reconnected")
_DEFAULT_DISCONNECTED_CB = partial(_log_nats_connection_event, event="disconnected")
_DEFAULT_CLOSED_CB = partial(_log_nats_connection_event, event="closed")


async def aget_nats_connect(
    servers: str | None = None,
    creds: str | None = None,
    connect_timeout: int = _CONNECTION_TIMEOUT,
    max_reconnect_attempts: int = _MAX_RECONNECT_ATTEMPTS,
    ping_interval: int = _PING_INTERVAL,
    max_outstanding_pings: int = _MAX_OUTSTANDING_PINGS,
    error_cb: Callable[[Exception], Awaitable[Any]] = _log_nats_error,
    reconnected_cb: Callable[[], Awaitable[Any]] = _DEFAULT_RECONNECTED_CB,
    disconnected_cb: Callable[[], Awaitable[Any]] = _DEFAULT_DISCONNECTED_CB,
    closed_cb: Callable[[], Awaitable[Any]] = _DEFAULT_CLOSED_CB,
) -> NATS:
    servers = servers or settings.nats_url
    assert servers is not None, "nats servers or settings.tfy_base_url is not set"
    if not creds:
        logger.info("Getting NATS credentials...")
        creds = await _get_nats_credentials()
    logger.info("Connecting to NATS...")
    nc = await nats.connect(
        servers=servers,
        user_credentials=UserString(creds),
        connect_timeout=connect_timeout,
        max_reconnect_attempts=max_reconnect_attempts,
        ping_interval=ping_interval,
        max_outstanding_pings=max_outstanding_pings,
        error_cb=error_cb,
        reconnected_cb=reconnected_cb,
        disconnected_cb=disconnected_cb,
        closed_cb=closed_cb,
    )
    logger.info("Connected to NATS")
    return nc


@asynccontextmanager
async def nats_connection() -> AsyncIterator[tuple[NATS, JetStreamContext]]:
    nc = await aget_nats_connect()
    js = JetStreamContext(conn=nc, timeout=JS_TIMEOUT)
    try:
        yield nc, js
    finally:
        await nc.drain()
        await nc.close()



import inspect
import os
import sys
from collections.abc import Callable
from enum import Enum, auto
from itertools import chain
from textwrap import dedent
from typing import Annotated, Any, Generic, Literal, TypeVar, Union

import logfire
from fastapi.concurrency import run_in_threadpool
from openai import NOT_GIVEN, NotGiven
from openai.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionToolParam,
)
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    create_model,
    model_validator,
)

from tfy_assistant_framework.settings import settings
from tfy_assistant_framework.swarm._logger import AgentLogLine, log

__CTX_VARS_NAME__ = "context_variables"
ContextVarType = dict[str, Any]


def get_obj_type_identity(obj: object) -> str:
    c = obj.__class__
    module: str | None = c.__module__
    if module == "__main__":
        module = sys.modules[module].__file__
        assert module is not None
        module = os.path.splitext(os.path.basename(module))[0]
    return f"{module}:{obj.__class__.__qualname__}"


class Tool(BaseModel):
    name: str
    func: Callable[..., Any]
    model: type[BaseModel]
    openai_function_schema: ChatCompletionToolParam
    pass_ctx: bool

    model_config = ConfigDict(arbitrary_types_allowed=True)

    async def call(
        self, args: dict[Any, Any], context_variables: ContextVarType
    ) -> Union[str, "Agent[Any]", dict[Any, Any], BaseModel]:
        kwargs = {}
        if self.pass_ctx:
            kwargs[__CTX_VARS_NAME__] = context_variables

        args_model = self.model(**args)
        keys = chain(
            self.model.model_fields.keys(), args_model.model_computed_fields.keys()
        )
        for k in keys:
            kwargs[k] = getattr(args_model, k)

        with logfire.span(
            f"Tool: {self.name}",
            attributes={
                "tool.name": self.name,
                "tool.args": kwargs,
                "tool.context_variables": context_variables,
            },
        ) as span:
            if inspect.iscoroutinefunction(self.func):
                result = await self.func(**kwargs)
            else:
                result = await run_in_threadpool(self.func, **kwargs)

            # Add result to span attributes
            span.set_attributes(
                {
                    "tool.result": result,
                }
            )
            return result

    @classmethod
    def from_function(cls, func: Callable[..., Any]) -> "Tool":
        try:
            signature = inspect.signature(func)
        except ValueError as e:
            raise ValueError(
                f"Failed to get signature for function {func.__name__}: {e!s}"
            ) from e

        parameters: dict[str, tuple[type, Any]] = {}
        pass_ctx = False
        for param_name, param in signature.parameters.items():
            # hide context_variables from model
            if param_name == __CTX_VARS_NAME__:
                pass_ctx = True
                continue
            if param.annotation == inspect.Parameter.empty:
                raise ValueError(
                    f"Type annotation required for function: {func.__name__} argument: {param.name}"
                )
            if param.kind in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.KEYWORD_ONLY,
            ):
                raise ValueError(
                    f"{param.kind} not supported for function: {func.__name__} argument: {param.name}"
                )
            parameters[param_name] = (
                param.annotation,
                param.default if param.default != inspect.Parameter.empty else ...,
            )

        model = create_model(func.__qualname__, **parameters)  # type: ignore
        return cls(
            name=func.__name__,
            func=func,
            model=model,
            pass_ctx=pass_ctx,
            openai_function_schema=ChatCompletionToolParam(
                {
                    "type": "function",
                    "function": {
                        "name": func.__name__,
                        "description": func.__doc__ or "",
                        "parameters": model.model_json_schema(),
                    },
                }
            ),
        )


class AgentResponseValidationError(BaseModel):
    error_message: Annotated[
        str,
        Field(min_length=1),
    ]


AgentResponseType = TypeVar("AgentResponseType", bound=BaseModel)


class Agent(BaseModel, Generic[AgentResponseType]):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str = ""
    model: str = settings.openai_model_name
    instructions: str | Callable[[dict[Any, Any] | None], str]
    functions: list[Callable[..., Any]] = []
    _tools: dict[str, Tool | ChatCompletionToolParam] = {}
    tool_choice: Literal["none", "auto", "required"] | NotGiven = NOT_GIVEN
    parallel_tool_calls: bool = False

    response_type: type[AgentResponseType] | None = None
    response_tool_name: Annotated[
        str,
        Field(min_length=1),
    ] = "RESPONSE"
    response_tool_description: Annotated[
        str,
        Field(min_length=1),
    ] = """
Call this tool once you have achieved your goal.
    """

    @model_validator(mode="after")
    def process_response_tool_description(self) -> "Agent[AgentResponseType]":
        self.response_tool_description = dedent(self.response_tool_description)
        return self

    @model_validator(mode="after")
    def set_name(self) -> "Agent[AgentResponseType]":
        self.name = self.name if self.name else get_obj_type_identity(self)
        return self

    @model_validator(mode="after")
    def create_tools(self) -> "Agent[AgentResponseType]":
        if self.tools:
            return self
        for f in self.functions:
            t = Tool.from_function(f)
            if t.name in self._tools:
                raise ValueError(f"Duplicate tool {t.name}")
            self._tools[t.name] = t
        if self.response_type:
            if self.response_tool_name in self._tools:
                raise ValueError(f"Duplicate tool {self.response_tool_name}")
            self._tools[self.response_tool_name] = ChatCompletionToolParam(
                {
                    "type": "function",
                    "function": {
                        "name": self.response_tool_name,
                        "description": self.response_tool_description,
                        "parameters": self.response_type.model_json_schema(),  # type: ignore
                    },
                }
            )
        return self

    @property
    def tools(self) -> dict[str, Tool | ChatCompletionToolParam]:
        return self._tools

    async def validate_response(
        self, response: AgentResponseType, context_variables: ContextVarType
    ) -> AgentResponseValidationError | AgentResponseType:
        return response


def create_structured_agent(
    name: str,
    response_type: type[AgentResponseType],
    instruction_template: str,
    model: str = settings.openai_model_name,
) -> Agent[AgentResponseType]:
    def generate_instructions(context_variables: ContextVarType | None) -> str:
        assert context_variables is not None
        return instruction_template.format(d=context_variables)

    return Agent(
        name=name,
        model=model,
        response_type=response_type,
        instructions=generate_instructions,
        tool_choice="required",
    )


class AgentRunState(Enum):
    EMPTY = auto()
    AI = auto()
    HUMAN = auto()
    AI_IN_TOOL_CALLS = auto()


ALLOWED_AGENT_RUN_STATE_TRANSITION: dict[AgentRunState, set[AgentRunState]] = {
    AgentRunState.EMPTY: {
        AgentRunState.AI,
        AgentRunState.HUMAN,
        AgentRunState.AI_IN_TOOL_CALLS,
    },
    AgentRunState.AI: {
        AgentRunState.HUMAN,
        AgentRunState.AI_IN_TOOL_CALLS,
        AgentRunState.AI,
    },
    AgentRunState.HUMAN: {AgentRunState.AI, AgentRunState.AI_IN_TOOL_CALLS},
    AgentRunState.AI_IN_TOOL_CALLS: {AgentRunState.AI, AgentRunState.AI_IN_TOOL_CALLS},
}


def message_to_state(
    m: ChatCompletionMessageParam, pending_tool_calls: set[str]
) -> AgentRunState:
    match m["role"]:
        case "user":
            return AgentRunState.HUMAN
        case "assistant":
            if m.get("tool_calls"):
                return AgentRunState.AI_IN_TOOL_CALLS
            return AgentRunState.AI
        case "tool":
            if pending_tool_calls:
                return AgentRunState.AI
            return AgentRunState.AI_IN_TOOL_CALLS
        case _:
            raise ValueError(f"Unknown ${m['role']}")


class AgentRun(BaseModel, Generic[AgentResponseType]):
    messages: list[ChatCompletionMessageParam] = []
    last_logged_message_index: int = 0
    agent: Agent[AgentResponseType]
    context_variables: ContextVarType = {}
    response: AgentResponseType | None = None
    state: AgentRunState = AgentRunState.EMPTY
    pending_tool_calls: set[str] = Field(default_factory=set)

    def update(
        self,
        message: ChatCompletionMessageParam,
        context_variables: ContextVarType | None = None,
        agent: Agent[AgentResponseType] | None = None,
    ) -> None:
        if message["role"] == "assistant":
            tool_calls = message.get("tool_calls")
            if tool_calls:
                self.pending_tool_calls = {t["id"] for t in tool_calls}
        if message["role"] == "tool":
            tool_call_id = message.get("tool_call_id")
            if tool_call_id:
                self.pending_tool_calls.discard(tool_call_id)
        transitioning_to_state = message_to_state(message, self.pending_tool_calls)
        if transitioning_to_state not in ALLOWED_AGENT_RUN_STATE_TRANSITION[self.state]:
            raise ValueError(
                f"{self.state!r} to {transitioning_to_state!r} not allowed"
            )

        self.state = transitioning_to_state
        self.messages.append(message)
        if context_variables:
            self.context_variables.update(context_variables)
        if agent:
            self.agent = agent

    def debug_log(self) -> None:
        while self.last_logged_message_index < len(self.messages):
            log(
                AgentLogLine(
                    agent_name=self.agent.name,
                    num_messages=self.last_logged_message_index + 1,
                    current_message=self.messages[self.last_logged_message_index],
                )
            )
            self.last_logged_message_index += 1

    def last_message(self) -> ChatCompletionMessageParam:
        return self.messages[-1]


class AgentFunctionResult(BaseModel):
    """
    Encapsulates the possible return values for an agent function.

    Attributes:
        value (str): The result value as a string.
        agent (Agent): The agent instance, if applicable.
        context_variables (dict): A dictionary of context variables.
    """

    value: str = ""
    agent: Agent[Any] | None = None
    context_variables: dict[Any, Any] = {}



from pydantic import BaseModel
from rich.console import Console
from rich.pretty import pprint

console = Console(soft_wrap=True)


async def _ask_user_from_console(question: str) -> str:
    """
    Ask user for input form console and return the response
    Args:
        question: Question to ask the user

    Returns:
        User input as string
    """
    return console.input(f"\n[bold cyan] ({question})[/]: ")


async def _fake_send_update_to_user(message: str) -> None:
    pprint(message)


async def _fake_send_streaming_update_to_user(update: BaseModel) -> None:
    pprint(update.model_dump_json())



import time
import uuid
from abc import ABC
from datetime import UTC, datetime
from enum import Enum, unique
from typing import Generic, Literal, TypeVar

from pydantic import (
    AwareDatetime,
    BaseModel,
    ConfigDict,
    Field,
    SerializeAsAny,
    computed_field,
)


@unique
class UserMessageTypes(str, Enum):
    NONE = "NONE"
    CHAT = "CHAT"


@unique
class AssistantState(str, Enum):
    WAITING_FOR_USER_INPUT = "WAITING_FOR_USER_INPUT"
    BUSY = "BUSY"
    TERMINATED = "TERMINATED"


class UserChatMessage(BaseModel):
    message_type: Literal[UserMessageTypes.CHAT]
    content: str


class IncomingMessage(BaseModel):
    message: UserChatMessage


RoleT = TypeVar("RoleT", bound=Literal["assistant", "user"])
RightSideContentT = TypeVar("RightSideContentT", bound=Literal["markdown"])
MessageTypeT = TypeVar("MessageTypeT", bound=UserMessageTypes)


class UpdateToUser(BaseModel, Generic[RoleT, MessageTypeT], ABC):
    model_config = ConfigDict(extra="allow")
    message_id: str = Field(default_factory=lambda: uuid.uuid1().hex)
    timestamp: AwareDatetime = Field(default_factory=lambda: datetime.now(UTC))
    content: str
    role: RoleT
    expected_user_message_type: MessageTypeT
    state_: AssistantState | None = Field(default=None, alias="state", exclude=True)

    # TODO: this is kept for compatibility with Frontend, remove it later, and use the content field directly
    @computed_field
    def display_message(self) -> str:
        return self.content

    @computed_field
    def state(self) -> AssistantState:
        if self.state_ is not None:
            return self.state_
        if (
            self.role == "assistant"
            and self.expected_user_message_type == UserMessageTypes.CHAT
        ):
            return AssistantState.WAITING_FOR_USER_INPUT
        return AssistantState.BUSY


class RightSideContent(BaseModel, Generic[RightSideContentT], ABC):
    type: RightSideContentT


class RightSideMarkdownContent(RightSideContent[Literal["markdown"]]):
    type: Literal["markdown"] = "markdown"
    title: str
    content: str


class AssistantUpdateToUser(UpdateToUser[Literal["assistant"], UserMessageTypes]):
    """
    Represents an update sent to the user from the assistant.

    This class extends `UpdateToUser` with specific attributes to define the role, expected user response type, and additional content elements.

    Attributes:
        role (Literal["assistant"]): Specifies the role of the sender as "assistant".
        expected_user_message_type (UserMessageTypes): Defines the expected type of user response.
            Defaults to `UserMessageTypes.NONE`, indicating no response is required.
        right_side_content (list[BaseModel] | None): Optional content displayed on the right side of the user interface.
        right_side_streaming_content (BaseModel | None): Optional streaming content displayed on the right side of the user interface.
    """

    role: Literal["assistant"] = "assistant"
    expected_user_message_type: UserMessageTypes = UserMessageTypes.NONE
    right_side_content: SerializeAsAny[list[BaseModel]] | None = None
    right_side_streaming_content: SerializeAsAny[BaseModel] | None = None


class UserUpdateToUser(UpdateToUser[Literal["user"], Literal[UserMessageTypes.NONE]]):
    role: Literal["user"] = "user"
    expected_user_message_type: Literal[UserMessageTypes.NONE] = UserMessageTypes.NONE


class Log(BaseModel):
    # This follows Loki's log format
    type: Literal["log"] = "log"
    log: str
    stream: str
    time: int = Field(default_factory=lambda: int(time.time_ns()))  # in nanoseconds



import httpx

from tfy_assistant_framework._logger import logger

async_http_client = httpx.AsyncClient()

REQUEST_TIMEOUT = 30


def log_and_raise_for_status(response: httpx.Response) -> None:
    """Log and raise an exception if the response status code is non-2xx."""
    if not response.is_success:
        logger.warning(
            "Request to %s %s failed with status code %d. Response body: %s.",
            response.request.method,
            response.request.url,
            response.status_code,
            response.text,
        )
    response.raise_for_status()



import logging


def setup_logger(name: str | None = None) -> logging.Logger:
    """Set up the logger for the module."""
    _logger = logging.getLogger(name)

    if not _logger.hasHandlers():
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        _logger.addHandler(handler)
        _logger.setLevel(logging.DEBUG)
    return _logger


# Set up the logger for the module
logger = setup_logger(__name__)



import json
from collections.abc import AsyncIterator, Awaitable
from contextlib import asynccontextmanager
from typing import Any

from fastapi import APIRouter, FastAPI, Response
from nats.aio.client import Client as NATS
from nats.errors import NoRespondersError
from pydantic import BaseModel

from tfy_assistant_framework.assistant._assistant_tasks import (
    AssistantTasks,
    CreateAssistantTaskRunResult,
)
from tfy_assistant_framework.assistant._base_assistant_io import AssistantUpdateSink
from tfy_assistant_framework.assistant._nats_assistant_io import (
    NATSAssistantInput,
    NATSInputResponse,
)


class PostAssistantTaskMessage(BaseModel):
    """Message model for sending data to an assistant task."""

    message: dict[str, Any] | None = None

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "message": {
                        "message": {"message_type": "CHAT", "content": "test message"}
                    }
                }
            ]
        }
    }


class AssistantServe:
    def __init__(
        self,
        path: str = "",
    ) -> None:
        self._assistant_tasks: AssistantTasks | None = None
        self._nc: NATS | None = None
        self._router = APIRouter(prefix=path)
        self._app: FastAPI | None = None

    @asynccontextmanager
    async def init(self, nc: NATS, app: FastAPI) -> AsyncIterator["AssistantServe"]:
        self._nc = nc
        self._assistant_tasks = AssistantTasks(nc=self._nc)
        self._app = app
        self._app.include_router(self._router)
        try:
            yield self
        finally:
            await self._assistant_tasks.cancel_tasks()
            self._assistant_tasks = None

    async def create_assistant_task(
        self,
        task_id: str,
        assistant_awaitable: Awaitable[Any],
        assistant_update_sink: AssistantUpdateSink,
    ) -> CreateAssistantTaskRunResult:
        assert self._assistant_tasks is not None
        return await self._assistant_tasks.create_assistant_task(
            task_id=task_id,
            assistant_awaitable=assistant_awaitable,
            assistant_update_sink=assistant_update_sink,
        )

    async def send_message_to_assistant_task(
        self,
        task_id: str,
        payload: PostAssistantTaskMessage,
    ) -> Response:
        sub = NATSAssistantInput.get_incoming_message_subject(task_id=task_id)
        assert self._nc is not None
        try:
            msg = await self._nc.request(
                sub, json.dumps(payload.message).encode(), timeout=10
            )
        except NoRespondersError:
            return Response(status_code=404, content=f"Task {task_id} does not exist")
        response = NATSInputResponse(**json.loads(msg.data.decode()))
        return Response(
            status_code=response.status_code, content=response.error_details
        )



import asyncio
import contextvars
from collections.abc import AsyncGenerator, Awaitable, Callable, Iterator
from contextlib import asynccontextmanager, contextmanager
from functools import partial
from typing import (
    Any,
    TypeVar,
)

import logfire
from fastapi import HTTPException
from pydantic import BaseModel

from tfy_assistant_framework._logger import logger
from tfy_assistant_framework.assistant._base_assistant_io import (
    AssistantUpdateSink,
)
from tfy_assistant_framework.assistant._console_assistant_io import (
    _ask_user_from_console,
    _fake_send_streaming_update_to_user,
    _fake_send_update_to_user,
)
from tfy_assistant_framework.assistant._types import (
    AssistantState,
    AssistantUpdateToUser,
    IncomingMessage,
    UpdateToUser,
    UserChatMessage,
    UserMessageTypes,
    UserUpdateToUser,
)
from tfy_assistant_framework.settings import settings

AssistantRequestBody = TypeVar(
    "AssistantRequestBody", bound=BaseModel | None, contravariant=True
)


_send_update_var: contextvars.ContextVar[
    Callable[[UpdateToUser[Any, Any]], Awaitable[None]] | None
] = contextvars.ContextVar("_send_update_var", default=None)


def _set_send_update_var(
    f: Callable[[UpdateToUser[Any, Any]], Awaitable[None]],
) -> None:
    assert _send_update_var.get() is None
    _send_update_var.set(f)


_active_assistant_runner: contextvars.ContextVar["AssistantRunner | None"] = (
    contextvars.ContextVar("_active_assistant_runner", default=None)
)


def _set_active_assistant_runner(
    f: "AssistantRunner",
) -> None:
    assert _active_assistant_runner.get() is None
    _active_assistant_runner.set(f)
    _set_send_update_var(f.send_update)


_streaming_update_ongoing: contextvars.ContextVar[bool | None] = contextvars.ContextVar(
    "_streaming_update_ongoing", default=None
)

_collected_updates_var: contextvars.ContextVar[list[BaseModel] | None] = (
    contextvars.ContextVar("_collected_updates_var", default=None)
)


def _unset_assistant_context_vars() -> None:
    """unset all context variables to their default values."""
    _active_assistant_runner.set(None)
    _send_update_var.set(None)
    _collected_updates_var.set(None)


@contextmanager
def collect_send_update() -> Iterator[list[BaseModel]]:
    try:
        data: list[BaseModel] = []
        _collected_updates_var.set(data)
        yield data
    finally:
        _collected_updates_var.set(None)


async def send_update(update: UpdateToUser[Any, Any]) -> None:
    collection = _collected_updates_var.get()
    if collection is not None:
        collection.append(update)
        return

    func = _send_update_var.get()
    if func is not None:
        await func(update)
        return

    await _fake_send_update_to_user(update.content)


def get_active_task_id() -> str | None:
    r = _active_assistant_runner.get()
    if r is not None:
        return r._task_id
    return None


@asynccontextmanager
async def init_streaming_update(
    assistant_update: AssistantUpdateToUser, right_side_content_title: str
) -> AsyncGenerator[Callable[[BaseModel], Awaitable[None]]]:
    """Initialize a streaming update context for the assistant update."""

    if _streaming_update_ongoing.get() is not None:
        raise RuntimeError(
            "init_streaming_update is already in progress. "
            "Nested init_streaming_update are not allowed. Ensure you are not calling init_streaming_update concurrently."
        )

    _streaming_update_ongoing.set(True)
    try:
        assistant_runner = _active_assistant_runner.get()
        collection = _collected_updates_var.get()
        if collection is not None:

            async def streaming_update_with_collection(content: BaseModel) -> None:
                collection.append(content)

            yield streaming_update_with_collection
        elif assistant_runner is not None:
            _assistant_update_sink = assistant_runner._update_sink
            assistant_update_streaming_content = (
                _assistant_update_sink.init_streaming_update(
                    parent_update=assistant_update,
                    title=right_side_content_title,
                )
            )
            await send_update(assistant_update_streaming_content)
            yield partial(
                _assistant_update_sink.send_streaming_update,
                parent_update=assistant_update,
            )
        elif settings.local:
            yield _fake_send_streaming_update_to_user
        else:
            raise Exception(
                "Please set env `local` to `True` to run the assistant locally."
            )
    finally:
        _streaming_update_ongoing.set(None)


async def get_user_input(prompt: str) -> str:
    assistant_runner = _active_assistant_runner.get()
    if assistant_runner is not None:
        return await assistant_runner.get_user_input(prompt)
    if settings.local:
        return await _ask_user_from_console(prompt)
    raise Exception("Please set env `local` to `True` to run the assistant locally.")


class AssistantRunner:
    def __init__(
        self,
        task_id: str,
        assistant_awaitable: Awaitable[None],
        assistant_update_sink: AssistantUpdateSink,
    ) -> None:
        self._task_id = task_id
        self._waiting_input_messages: list[str] = []
        self._expected_user_message_type: UserMessageTypes = UserMessageTypes.NONE
        self._assistant_awaitable = assistant_awaitable
        self._update_sink = assistant_update_sink
        _set_active_assistant_runner(self)

    async def send_update(self, update: UpdateToUser[Any, Any]) -> None:
        with logfire.span(
            "send_update",
            update=update,
        ):
            self._expected_user_message_type = update.expected_user_message_type
            await self._update_sink.send_update(update)

    async def get_user_input(self, prompt: str) -> str:
        with logfire.span(
            "assistant.get_user_input",
            prompt=prompt,
        ) as span:
            await self.send_update(
                AssistantUpdateToUser(
                    content=prompt,
                    expected_user_message_type=UserMessageTypes.CHAT,
                ),
            )

            while not self._waiting_input_messages:
                await asyncio.sleep(0.5)

            input_message = self._waiting_input_messages.pop()
            span.set_attributes(
                {
                    "user_input": input_message,
                }
            )

            await self.send_update(
                UserUpdateToUser(
                    content=input_message,
                ),
            )
            return input_message

    async def handle_incoming_message(self, body: IncomingMessage) -> None:
        if (
            body is not None
            and isinstance(body.message, UserChatMessage)
            and self._expected_user_message_type == UserMessageTypes.CHAT
        ):
            if len(self._waiting_input_messages) > 1:
                raise HTTPException(
                    status_code=400,
                    detail="Multiple user messages received while waiting for a single message",
                )
            self._waiting_input_messages.append(body.message.content)

    async def run(self) -> None:
        try:
            logger.info("Task %s started", self._task_id)

            await self._assistant_awaitable

            await send_update(
                AssistantUpdateToUser(
                    content="Assistant completed.",
                    state=AssistantState.TERMINATED,
                )
            )
            logger.info("Task %s completed", self._task_id)

        except asyncio.CancelledError:
            await send_update(
                AssistantUpdateToUser(
                    content="Assistant cancelled.",
                    state=AssistantState.TERMINATED,
                )
            )
            logger.info("Task %s cancelled", self._task_id)
        except Exception:
            logger.exception("Task %s failed with exception", self._task_id)
            await send_update(
                AssistantUpdateToUser(
                    content="Assistant failed.",
                    state=AssistantState.TERMINATED,
                )
            )
            raise
        finally:
            # unset all context variables to their defaults
            _unset_assistant_context_vars()



from collections.abc import Iterable
from datetime import datetime
from typing import cast

from openai.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCallParam,
)
from pydantic import BaseModel, Field, SkipValidation
from rich.console import Console, Group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text

console = Console()


class AgentLogLine(BaseModel):
    agent_name: str
    num_messages: int

    # https://github.com/pydantic/pydantic/issues/9467
    current_message: SkipValidation[ChatCompletionMessageParam]
    timestamp: datetime = Field(default_factory=lambda: datetime.now().astimezone())


def log(log_line: AgentLogLine) -> None:
    timestamp = log_line.timestamp.strftime("%Y-%m-%d %H:%M:%S")
    code_theme = "github-dark"
    role = log_line.current_message["role"]
    message_content = str(log_line.current_message.get("content") or "")
    tool_calls: Iterable[ChatCompletionMessageToolCallParam] = cast(
        list[ChatCompletionMessageToolCallParam],
        log_line.current_message.get("tool_calls", []),
    )

    rendered_content: list[Markdown | Text] = []

    if bool(message_content.strip()):
        rendered_content.append(Markdown(markup=message_content, code_theme=code_theme))

    role_str = (role or "Unknown").upper()

    for call in tool_calls:
        name = call["function"]["name"]
        args = call["function"]["arguments"]

        if name == "RESPONSE":
            rendered_content.append(
                Text.from_markup("[bold green]Response:[/bold green]")
            )
            rendered_content.append(
                Markdown(
                    markup=f"```json\n{args}\n```",
                    code_theme=code_theme,
                )
            )
        else:
            rendered_content.append(
                Text.from_markup("[bold magenta]Tool Calls:[/bold magenta]\n")
            )
            rendered_content.append(
                Markdown(
                    markup=f"```python\n▶ {name}({args})\n```", code_theme=code_theme
                )
            )

    if not rendered_content:
        return

    header = Text()
    header.append("Agent: ", style="bold cyan")
    header.append(log_line.agent_name, style="bold green")
    header.append(" | ", style="bold cyan")
    header.append(role_str, style="bold magenta")
    header.append(" | Total Messages: ", style="bold cyan")
    header.append(str(log_line.num_messages), style="bold yellow")
    header.append(" | Time: ", style="bold cyan")
    header.append(timestamp, style="bold blue")

    panel = Panel(
        Group(*rendered_content),
        title=header,
        border_style="bright_blue",
        padding=(1, 2),
    )

    console.print(panel)



from abc import abstractmethod
from collections.abc import Awaitable, Callable
from types import TracebackType
from typing import Any, Protocol

from pydantic import BaseModel

from tfy_assistant_framework.assistant._types import IncomingMessage, UpdateToUser


class AssistantInput(Protocol):
    _incoming_message_handler: Callable[[IncomingMessage], Awaitable[None]] | None

    @property
    def incoming_message_handler(self) -> Callable[[IncomingMessage], Awaitable[None]]:
        assert self._incoming_message_handler is not None
        return self._incoming_message_handler

    async def __aenter__(self) -> "AssistantInput": ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None: ...


class AssistantUpdateSink(Protocol):
    @abstractmethod
    async def send_update(self, update: UpdateToUser[Any, Any]) -> None: ...

    @abstractmethod
    def init_streaming_update(
        self, parent_update: UpdateToUser[Any, Any], title: str
    ) -> UpdateToUser[Any, Any]: ...

    @abstractmethod
    async def send_streaming_update(
        self, update: BaseModel, parent_update: UpdateToUser[Any, Any]
    ) -> None: ...

    async def get_sink_config(self) -> dict[Any, Any] | None:
        return None



"""Settings for the Framework."""

import os
from urllib.parse import urlparse

from dotenv import load_dotenv
from pydantic import ValidationInfo, field_validator
from pydantic_settings import BaseSettings

# Load environment variables from .env file
load_dotenv(dotenv_path=os.getenv("ENV_PATH", ".env"))


class Settings(BaseSettings):
    """Settings for the Framework."""

    local: bool = False
    openai_api_key: str
    openai_base_url: str
    openai_model_name: str

    # TrueFoundry settings
    tfy_base_url: str | None = None
    tfy_api_key: str | None = None
    tfy_assumed_user: str | None = None
    tfy_log_to_gateway: bool = True

    # NATS settings
    nats_url: str | None = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.tfy_base_url = (
            self.tfy_base_url.rstrip("/") if self.tfy_base_url else self.tfy_base_url
        )

    @field_validator("nats_url")
    @classmethod
    def generate_tfy_nats_url(
        cls, value: str | None, info: ValidationInfo
    ) -> str | None:
        """Validate the NATS URL."""
        if value:
            return value

        tfy_base_url = info.data["tfy_base_url"]
        if tfy_base_url:
            parsed_url = urlparse(tfy_base_url)
            return f"wss://{parsed_url.hostname}/"

        # Return None as nats_url is not mandatory for all agents
        return None

    @property
    def tfy_host(self) -> str | None:
        if self.tfy_base_url:
            parsed_url = urlparse(self.tfy_base_url)
            return f"{parsed_url.scheme}://{parsed_url.netloc}"
        return None


settings = Settings()



# https://github.com/openai/swarm/blob/9db581cecaacea0d46a933d6453c312b034dbf47/swarm/__init__.py#L1
from ._core import (
    MaxTurnsExceeded,
    Swarm,
    client,
    get_response_from_structured_agent,
)
from ._types import (
    Agent,
    AgentResponseValidationError,
    AgentRun,
    AgentRunState,
    ContextVarType,
    create_structured_agent,
)

__all__ = [
    "Agent",
    "AgentResponseValidationError",
    "AgentRun",
    "AgentRunState",
    "ContextVarType",
    "MaxTurnsExceeded",
    "Swarm",
    "client",
    "create_structured_agent",
    "get_response_from_structured_agent",
]



import copy
import json
import logging
import uuid
from collections.abc import AsyncIterable
from typing import Any

import logfire
from openai import NOT_GIVEN, AsyncOpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCall,
    ChatCompletionMessageToolCallParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import Function
from pydantic import BaseModel, ValidationError

from tfy_assistant_framework.assistant._assistant_runner import (
    get_active_task_id,
    get_user_input,
)
from tfy_assistant_framework.settings import settings
from tfy_assistant_framework.swarm._types import (
    Agent,
    AgentFunctionResult,
    AgentResponseType,
    AgentResponseValidationError,
    AgentRun,
    AgentRunState,
    ContextVarType,
    Tool,
)

logger = logging.getLogger(__name__)
__CTX_VARS_NAME__ = "context_variables"


class Swarm:
    def __init__(
        self,
        client: AsyncOpenAI | None = None,
    ) -> None:
        if not client:
            client = AsyncOpenAI()
        self.client = client
        logfire.instrument_openai(client)

    async def get_chat_completion(
        self,
        agent: Agent[Any],
        history: list[ChatCompletionMessageParam],
        context_variables: dict[Any, Any],
        model_override: str | None,
        tracking_session_id: str | None,
        max_tokens: int | None = None,
    ) -> ChatCompletion:
        instructions = (
            agent.instructions(context_variables)
            if callable(agent.instructions)
            else agent.instructions
        )
        system_prompt = ChatCompletionSystemMessageParam(
            content=instructions, role="system"
        )
        # how do we log the system prompt?
        messages = [system_prompt, *history]

        tools = [
            t.openai_function_schema if isinstance(t, Tool) else t
            for t in agent.tools.values()
        ]

        tfy_metadata = {
            "tfy_log_request": "true" if settings.tfy_log_to_gateway else "false",
        }
        if tracking_session_id:
            tfy_metadata["session_id"] = tracking_session_id
        tfy_metadata["agent_name"] = agent.name

        return await self.client.chat.completions.create(
            model=model_override or agent.model,
            messages=messages,
            tools=tools or NOT_GIVEN,
            tool_choice=agent.tool_choice,
            parallel_tool_calls=agent.parallel_tool_calls if tools else NOT_GIVEN,
            extra_headers={"X-TFY-METADATA": json.dumps(tfy_metadata)},
            max_tokens=max_tokens,
        )

    def handle_function_result(
        self, result: str | Agent[Any] | dict[Any, Any] | BaseModel
    ) -> AgentFunctionResult:
        # Disable multi-agent for now.
        match result:
            case AgentFunctionResult() as result:
                assert result.agent is None, "multi-agent is unsupported"
                return result

            case Agent():
                raise AssertionError("multi-agent is unsupported")
                # return AgentFunctionResult(
                #     value=json.dumps({"assistant": agent.name}),
                #     agent=agent,
                # )
            case _:
                try:
                    return AgentFunctionResult(value=str(result))
                except Exception as e:
                    error_message = f"Failed to cast response to string: {result}. Make sure agent functions return a string or Result object. Error: {e!s}"

                    raise TypeError(error_message) from e

    async def handle_tool_calls(
        self,
        agent: Agent[Any],
        tool_calls: list[ChatCompletionMessageToolCall],
        agent_run: AgentRun[Any],
    ) -> AsyncIterable[None]:
        for tool_call in tool_calls:
            name = tool_call.function.name
            args = json.loads(tool_call.function.arguments)
            if agent.response_type and name == agent.response_tool_name:
                if len(tool_calls) > 1:
                    agent_run.update(
                        message=ChatCompletionToolMessageParam(
                            content=f"Error: Tool {name} cannot be used with other tool calls.",
                            role="tool",
                            tool_call_id=tool_call.id,
                        )
                    )
                    yield
                    continue

                try:
                    r = agent.response_type(**args)
                except ValidationError as ex:
                    agent_run.update(
                        message=ChatCompletionToolMessageParam(
                            content=f"Fix the arguments to resolve the validation error in the tool call.\nError: {ex}",
                            role="tool",
                            tool_call_id=tool_call.id,
                        )
                    )
                    yield
                    continue

                validation_result = await agent.validate_response(
                    r, agent_run.context_variables
                )
                match validation_result:
                    case agent.response_type() as validation_result:  # type: ignore
                        agent_run.response = validation_result
                        return
                    case AgentResponseValidationError() as validation_result:
                        agent_run.update(
                            message=ChatCompletionToolMessageParam(
                                content=validation_result.error_message,
                                role="tool",
                                tool_call_id=tool_call.id,
                            )
                        )
                        yield
                        continue
                    case _:
                        raise TypeError(f"Unknown type {type(validation_result)}")

            # handle missing tool case, skip to next tool
            if name not in agent.tools:
                agent_run.update(
                    message=ChatCompletionToolMessageParam(
                        content=f"Error: Tool {name} not found.",
                        role="tool",
                        tool_call_id=tool_call.id,
                    )
                )
                yield
                continue

            tool = agent.tools[name]
            raw_result = None
            try:
                assert isinstance(tool, Tool)
                raw_result = await tool.call(args, agent_run.context_variables)
            except ValidationError as ex:
                agent_run.update(
                    message=ChatCompletionToolMessageParam(
                        content=f"Fix the arguments to resolve the validation error in the tool call.\nError: {ex}",
                        role="tool",
                        tool_call_id=tool_call.id,
                    )
                )
                yield
                continue
            except Exception as ex:
                logger.warning("Error calling tool %s: %s", name, ex, exc_info=True)
                agent_run.update(
                    message=ChatCompletionToolMessageParam(
                        content=f"Error: {ex}",
                        role="tool",
                        tool_call_id=tool_call.id,
                    )
                )
                yield
                continue

            result: AgentFunctionResult = self.handle_function_result(raw_result)
            agent_run.update(
                message=ChatCompletionToolMessageParam(
                    content=result.value,
                    role="tool",
                    tool_call_id=tool_call.id,
                ),
                context_variables=result.context_variables,
                agent=result.agent,
            )
            yield

    async def run(
        self,
        agent: Agent[Any],
        messages: list[ChatCompletionMessageParam],
        context_variables: ContextVarType | None = None,
        model_override: str | None = None,
        # no int inf in python
        max_turns: float | int = float("inf"),
        tracking_session_id: str | None = None,
        interactive: bool = False,
        max_tokens: int | None = None,
    ) -> AsyncIterable[AgentRun[Any]]:
        with logfire.span(
            f"Agent: {agent.name}",
            agent=agent,
            context_variables=context_variables,
            max_turns=max_turns,
            interactive=interactive,
            messages=messages,
        ) as span:
            context_variables = context_variables or {}
            messages = copy.deepcopy(messages)
            agent_run: AgentRun[Any] = AgentRun(agent=agent)
            span.set_attributes(
                {
                    "agent.name": agent.name,
                    "max_turns": max_turns,
                    "interactive": interactive,
                    "context_variables": context_variables,
                }
            )
            yield agent_run
            for m in messages:
                agent_run.update(
                    message=m,
                    context_variables=context_variables,
                    agent=agent,
                )
                yield agent_run

            init_len = len(messages)
            tracking_session_id = (
                tracking_session_id or get_active_task_id() or uuid.uuid4().hex
            )

            while len(agent_run.messages) - init_len < max_turns:
                # get completion with current history, agent
                completion = await self.get_chat_completion(
                    agent=agent_run.agent,
                    history=agent_run.messages,
                    context_variables=context_variables,
                    model_override=model_override,
                    tracking_session_id=tracking_session_id,
                    max_tokens=max_tokens,
                )
                choice_message = completion.choices[0].message
                message = ChatCompletionAssistantMessageParam(
                    role=choice_message.role,
                    content=choice_message.content,
                    refusal=choice_message.refusal,
                )
                if choice_message.tool_calls:
                    message["tool_calls"] = [
                        ChatCompletionMessageToolCallParam(
                            id=tc.id,
                            type=tc.type,
                            function=Function(
                                name=tc.function.name,
                                arguments=tc.function.arguments,
                            ),
                        )
                        for tc in choice_message.tool_calls
                    ]
                agent_run.update(
                    message=message,
                    context_variables=context_variables,
                    agent=agent,
                )
                yield agent_run
                # handle function calls, updating context_variables, and switching agents
                async for _ in self.handle_tool_calls(
                    agent,
                    choice_message.tool_calls or [],
                    agent_run=agent_run,
                ):
                    yield agent_run
                if agent_run.response:
                    span.set_attributes({"agent.response": agent_run.response})
                    yield agent_run
                    return

                if (
                    agent_run.state == AgentRunState.AI
                ):  # if llm responds without a tool call or response model it will get stuck
                    if interactive:  # If fetch_input_from_user is provided, then we assume that user input is required to proceed
                        assert "content" in agent_run.messages[-1]
                        message_content = agent_run.messages[-1]["content"]
                        prompt = (
                            str(message_content) if message_content is not None else ""
                        )
                        user_input = await get_user_input(prompt)
                        init_len = len(agent_run.messages)
                        agent_run.update(
                            message=ChatCompletionUserMessageParam(
                                content=user_input, role="user"
                            ),
                            context_variables=context_variables,
                            agent=agent,
                        )
                        yield agent_run
                    elif (
                        agent_run.agent.response_type is not None
                    ):  # If fetch_input_from_user is not provided and response_type is not None, then we force the agent to use only the tools available to it instead of responding to user
                        logger.warning(
                            "User input method not provided and agent is expected to return a response model, forcing agent to use only the tools available to it instead of responding to user"
                        )
                        agent_run.update(
                            message=ChatCompletionUserMessageParam(
                                content="YOU ARE ONLY ALLOWED TO USE THE TOOLS AVAILABLE TO YOU AND RESPOND WITH THE TOOL CALLS",
                                role="user",
                            ),
                            context_variables=context_variables,
                            agent=agent,
                        )
                        yield agent_run


client = Swarm()


class MaxTurnsExceeded(Exception): ...


async def get_response_from_structured_agent(
    agent: Agent[AgentResponseType],
    context_variables: ContextVarType | None = None,
    tracking_session_id: str | None = None,
    messages: list[ChatCompletionMessageParam] | None = None,
    interactive: bool = False,
    max_turns: int = 20,
    max_tokens: int | None = None,
) -> AgentResponseType:
    agent_run = None
    async for agent_run in client.run(
        agent=agent,
        messages=messages or [],
        interactive=interactive,
        context_variables=context_variables,
        tracking_session_id=tracking_session_id,
        max_turns=max_turns,
        max_tokens=max_tokens,
    ):
        agent_run.debug_log()
    assert agent_run is not None

    if agent_run.response is None:
        raise MaxTurnsExceeded()

    return agent_run.response



import json
from collections.abc import Awaitable, Callable
from types import TracebackType
from typing import Any, Literal

from fastapi import HTTPException
from nats.aio.client import Client as NATS
from nats.aio.msg import Msg
from nats.aio.subscription import Subscription
from nats.js import JetStreamContext
from pydantic import BaseModel

from tfy_assistant_framework._logger import logger
from tfy_assistant_framework.assistant._base_assistant_io import (
    AssistantInput,
    AssistantUpdateSink,
)
from tfy_assistant_framework.assistant._types import (
    AssistantUpdateToUser,
    IncomingMessage,
    UpdateToUser,
    UserMessageTypes,
)
from tfy_assistant_framework.nats_client import JS_PUBLISH_TIMEOUT

_OUTGOING_MESSAGE_STREAM_NAME = "llm-agent-task-state"
_INCOMING_MESSAGE_SUBJECT_NAME = "llm-agent-task-input"


class NATSInputResponse(BaseModel):
    status_code: int
    error_details: str | None = None


class RightSideStreamingUpdate(BaseModel):
    type: Literal["nats-js"] = "nats-js"
    title: str
    stream: str
    subject: str


class NATSAssistantInput(AssistantInput):
    def __init__(
        self,
        task_id: str,
        nc: NATS,
        incoming_message_handler: Callable[[IncomingMessage], Awaitable[None]],
    ) -> None:
        self._task_id = task_id
        self._nc = nc
        self._sub: Subscription | None = None
        self._incoming_message_handler = incoming_message_handler

    @staticmethod
    def get_incoming_message_subject(
        task_id: str,
    ) -> str:
        return f"{_INCOMING_MESSAGE_SUBJECT_NAME}.{task_id}"

    async def _handle_msg(self, msg: Msg) -> None:
        try:
            data = json.loads(msg.data)
            incoming_message = IncomingMessage(**data)
            await self.incoming_message_handler(incoming_message)
            response = NATSInputResponse(status_code=200)
        except HTTPException as ex:
            response = NATSInputResponse(
                status_code=ex.status_code, error_details=ex.detail
            )
        except Exception as ex:
            logger.exception(
                "Task %s failed to process incoming message", self._task_id
            )
            response = NATSInputResponse(status_code=500, error_details=str(ex))

        await self._nc.publish(msg.reply, response.model_dump_json().encode())

    async def __aenter__(self) -> "AssistantInput":
        subject = self.get_incoming_message_subject(task_id=self._task_id)
        logger.info("subscribing to %s for inputs", subject)
        self._sub = await self._nc.subscribe(
            subject=subject,
            cb=self._handle_msg,
        )
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._sub:
            try:
                await self._sub.unsubscribe()
            except Exception:
                logger.exception("Task %s failed unsubscribe IO", self._task_id)
        self._sub = None


class NATSAssistantUpdateSink(AssistantUpdateSink):
    def __init__(self, task_id: str, js: JetStreamContext) -> None:
        self._task_id = task_id
        self._js = js

    def _get_outgoing_update_subject(self) -> str:
        return f"{_OUTGOING_MESSAGE_STREAM_NAME}.{self._task_id}"

    def _get_streaming_update_subject(
        self,
        parent_update_id: str,
    ) -> str:
        return f"{_OUTGOING_MESSAGE_STREAM_NAME}.{self._task_id}.{parent_update_id}"

    async def send_update(self, update: UpdateToUser[Any, Any]) -> None:
        payload_json = update.model_dump(mode="json")
        await self._js.publish(
            subject=self._get_outgoing_update_subject(),
            payload=json.dumps(payload_json).encode("utf-8"),
            stream=_OUTGOING_MESSAGE_STREAM_NAME,
            timeout=JS_PUBLISH_TIMEOUT,
        )

    async def get_sink_config(self) -> dict[Any, Any] | None:
        return {
            "type": "nats-js",
            "subject": self._get_outgoing_update_subject(),
            "stream": _OUTGOING_MESSAGE_STREAM_NAME,
        }

    def init_streaming_update(
        self, parent_update: UpdateToUser[Any, Any], title: str
    ) -> AssistantUpdateToUser:
        assert isinstance(parent_update, AssistantUpdateToUser)
        assert parent_update.expected_user_message_type is UserMessageTypes.NONE
        parent_update.right_side_streaming_content = RightSideStreamingUpdate(
            title=title,
            stream=_OUTGOING_MESSAGE_STREAM_NAME,
            subject=self._get_streaming_update_subject(
                parent_update_id=parent_update.message_id,
            ),
        )
        return parent_update

    async def send_streaming_update(
        self,
        update: BaseModel,
        parent_update: UpdateToUser[Any, Any],
    ) -> None:
        payload_json = update.model_dump(mode="json")
        await self._js.publish(
            subject=self._get_streaming_update_subject(
                parent_update_id=parent_update.message_id,
            ),
            payload=json.dumps(payload_json).encode("utf-8"),
            stream=_OUTGOING_MESSAGE_STREAM_NAME,
            timeout=JS_PUBLISH_TIMEOUT,
        )


class TruefoundryNATSAssistantUpdateSink(NATSAssistantUpdateSink):
    def __init__(self, task_id: str, js: JetStreamContext, tenant_name: str) -> None:
        super().__init__(task_id, js)
        self._tenant_name = tenant_name

    def _get_outgoing_update_subject(self) -> str:
        return f"{_OUTGOING_MESSAGE_STREAM_NAME}.{self._tenant_name}.{self._task_id}"

    def _get_streaming_update_subject(
        self,
        parent_update_id: str,
    ) -> str:
        return f"{_OUTGOING_MESSAGE_STREAM_NAME}.{self._tenant_name}.{self._task_id}.{parent_update_id}"



from ._assistant_runner import (
    collect_send_update,
    get_active_task_id,
    get_user_input,
    init_streaming_update,
    send_update,
)
from ._assistant_tasks import CreateAssistantTaskRunResult
from ._base_assistant_io import AssistantUpdateSink
from ._nats_assistant_io import (
    NATSAssistantInput,
    NATSAssistantUpdateSink,
    NATSInputResponse,
    TruefoundryNATSAssistantUpdateSink,
)
from ._serve import AssistantServe, PostAssistantTaskMessage
from ._types import (
    AssistantState,
    AssistantUpdateToUser,
    Log,
    RightSideMarkdownContent,
    UserUpdateToUser,
)

__all__ = [
    "AssistantServe",
    "AssistantState",
    "AssistantUpdateSink",
    "AssistantUpdateToUser",
    "CreateAssistantTaskRunResult",
    "Log",
    "NATSAssistantInput",
    "NATSAssistantUpdateSink",
    "NATSInputResponse",
    "PostAssistantTaskMessage",
    "RightSideMarkdownContent",
    "TruefoundryNATSAssistantUpdateSink",
    "UserUpdateToUser",
    "collect_send_update",
    "get_active_task_id",
    "get_user_input",
    "init_streaming_update",
    "send_update",
]



import asyncio
from collections.abc import Awaitable
from dataclasses import dataclass
from typing import Any

from nats.aio.client import Client as NATS

from tfy_assistant_framework._logger import logger
from tfy_assistant_framework.assistant._assistant_runner import (
    AssistantRunner,
)
from tfy_assistant_framework.assistant._base_assistant_io import (
    AssistantInput,
    AssistantUpdateSink,
)
from tfy_assistant_framework.assistant._nats_assistant_io import NATSAssistantInput


@dataclass
class CreateAssistantTaskRunResult:
    task_id: str
    sink_config: dict[Any, Any]


_TASK_PRUNE_TIMEOUT_SECONDS = 30 * 60  # 30 minutes


class AssistantTasks:
    def __init__(self, nc: NATS) -> None:
        self._tasks: dict[str, tuple[asyncio.Task[None], AssistantRunner]] = {}
        self._nc = nc

    async def _run_task(
        self,
        assistant_runner: AssistantRunner,
        assistant_input: AssistantInput,
    ) -> None:
        async with assistant_input:
            await assistant_runner.run()

    async def create_assistant_task(
        self,
        task_id: str,
        assistant_awaitable: Awaitable[Any],
        assistant_update_sink: AssistantUpdateSink,
    ) -> CreateAssistantTaskRunResult:
        assistant_runner = AssistantRunner(
            task_id=task_id,
            assistant_awaitable=assistant_awaitable,
            assistant_update_sink=assistant_update_sink,
        )
        assistant_input = NATSAssistantInput(
            task_id=task_id,
            nc=self._nc,
            incoming_message_handler=assistant_runner.handle_incoming_message,
        )
        loop = asyncio.get_running_loop()
        task = loop.create_task(self._run_task(assistant_runner, assistant_input))
        self._tasks[task_id] = (task, assistant_runner)
        task.add_done_callback(lambda _: self._handle_task_completion(task_id=task_id))
        loop.call_later(_TASK_PRUNE_TIMEOUT_SECONDS, self._handle_task_timeout, task_id)
        sink_config = await assistant_update_sink.get_sink_config()
        if sink_config is None:
            raise ValueError("Sink config is required to create a task")
        return CreateAssistantTaskRunResult(
            task_id=task_id,
            sink_config=sink_config,
        )

    def _handle_task_timeout(self, task_id: str) -> None:
        task_tuple = self._tasks.get(task_id)
        if not task_tuple:
            logger.warning("Task %s not found for timeout", task_id)
            return
        task, _ = task_tuple
        task.cancel()
        logger.warning("Task %s was timed out", task_id)

    def _handle_task_completion(self, task_id: str) -> None:
        if task_id not in self._tasks:
            logger.warning("Task %s not found", task_id)
            return

        task_run, _ = self._tasks.pop(task_id)
        if task_run.cancelled():
            logger.info("Task %s was cancelled", task_id)
            return

        exception = task_run.exception()
        if exception:
            logger.error("Task %s failed with exception", task_id, exc_info=exception)
        else:
            logger.info("Task %s was successfully completed", task_id)

    async def cancel_tasks(self) -> None:
        # Create a list of tasks to avoid modifying dict during iteration
        tasks_to_cancel = list(self._tasks.items())

        for task_id, (task, _) in tasks_to_cancel:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.info("Task %s cancelled during shutdown", task_id)
            except Exception:
                logger.exception("Task %s failed during shutdown", task_id)
        self._tasks.clear()
